export const NAMESPACE_SVG = 'http://www.w3.org/2000/svg'
export const NAMESPACE_XHTML = 'http://www.w3.org/1999/xhtml'
export const NAMESPACE_XLINK = 'http://www.w3.org/1999/xlink'

const /** @type {Record<string, string>} */ NAMESPACE_MAP = {
  svg: NAMESPACE_SVG,
  xhtml: NAMESPACE_XHTML,
  xlink: NAMESPACE_XLINK
}

/**
 * A helper returning constructor class of given `arg`.
 */
export const typeOf = (/** @type {any} */ arg) => arg?.constructor

/**
 * A helper that checks if the given argument is of type `string`.
 *
 * @returns {arg is string}
 */
export const isString = (/** @type {any} */ arg) => typeOf(arg) === String

/**
 * A helper that checks if the given argument is of type `number`.
 *
 * @returns {arg is number}
 */
export const isNumber = (/** @type {any} */ arg) => typeOf(arg) === Number

/**
 * A helper that checks if the given argument is of type `Record<string, any>`.
 *
 * @returns {arg is Record<string, any>}
 */
export const isObject = (/** @type {any} */ arg) => typeOf(arg) === Object

/**
 * A helper that checks if the given argument is of type `Element`.
 *
 * @returns {arg is Element}
 */
export const isElement = (/** @type {any} */ arg) => arg?.nodeType === document.ELEMENT_NODE

/**
 * Replacement for the `in` operator (not to be confused with the `for-in` loop) that works properly.
 */
export const hasOwnProperty = (/** @type {any} */ ref, /** @type {any} */ key) =>
  (isString(key) || isNumber(key)) && Object.hasOwnProperty.call(ref ?? Object, key)

/**
 * A helper for convenient accessing to nested properties.
 */
export const get = (/** @type {any} */ ref, /** @type {string[]} */...keys) => {
  for (const key of keys) {
    if (!hasOwnProperty(ref, key)) {
      return
    }
    ref = ref[key]
  }
  return ref
}

/**
 * A helper that checks the equality of the given arguments.
 *
 * @returns {boolean}
 */
export const deepEqual = (/** @type {any} */ x, /** @type {any} */ y) => {
  if (x === y) {
    return true
  }

  const typeOfX = typeOf(x)

  if (typeOfX === typeOf(y)) {
    if (typeOfX === Date) {
      return +x === +y
    }

    if (typeOfX === String || typeOfX === RegExp) {
      return '' + x === '' + y
    }

    if (typeOfX === Array) {
      return x.length === y.length && x.every((/** @type {any} */ item, /** @type {number} */ index) => deepEqual(item, y[index]))
    }

    if (typeOfX === Object) {
      const keysOfX = Object.keys(x)
      return keysOfX.length === Object.keys(y).length && keysOfX.every(key => deepEqual(x[key], y[key]))
    }

    if (typeOfX === Set || typeOfX === Map) {
      if (x.size !== y.size) {
        return false
      }
      const xa = [...x]
      const ya = [...y]
      return xa.every(xv => ya.find(yv => deepEqual(xv, yv)))
    }
  }

  // eslint-disable-next-line no-self-compare
  return x !== x && y !== y
}

/**
 * The `$Config` format describes objects of type `Element`.
 *
 * - `t`: _tag_ of the created element
 * - `n`: modified _node_
 * - `a`: _attributes_ of the created or modified element set by `setAttribute` or `setAttributeNS`
 * - `p`: _properties_ to set in the created or modified element
 * - `i`: _items_ to append of the created or modified element
 *
 * @typedef {{
 *   t?: keyof HTMLElementTagNameMap | keyof SVGElementTagNameMap;
 *   n?: Node | string | number;
 *   a?: Record<string, string | number | boolean>;
 *   p?: Record<string, any>;
 *   i?: (Node | string | number | $Config)[];
 * }} $Config
 */

/**
 * @template T
 * @typedef {{
 *   t: T;
 * }} TConfig
 */

/**
 * @template N
 * @typedef {{
 *   n: N;
 * }} NConfig
 */

/**
 * ExtJS like, lightweight helper for creating and modifying DOM elements.
 *
 * @type {{
 *   <H extends keyof HTMLElementTagNameMap>(config: $Config & TConfig<H>, namespace?: string): HTMLElementTagNameMap[H];
 *   <S extends keyof  SVGElementTagNameMap>(config: $Config & TConfig<S>, namespace?: string):  SVGElementTagNameMap[S];
 *   <N extends HTMLElement | SVGElement>   (config: $Config & NConfig<N>, namespace?: string): N;
 * }}
 */
export const $ = (
  /** @type {$Config} */ { t, n, a = {}, p = {}, i = [] } = {},
  /** @type {string | null} */ namespace = null
) => {
  if (isString(n) || isNumber(n)) {
    // @ts-expect-error: TS2345
    n = document.createTextNode(n)
  }

  if (namespace == null && t === 'svg') {
    namespace = NAMESPACE_SVG
  }

  if (n == null && t != null) {
    n = namespace != null
      ? document.createElementNS(namespace, t)
      : document.createElement(t)
  }

  if (isElement(n)) {
    for (const attribute in a) {
      const value = a[attribute]
      const indexOfColon = attribute.indexOf(':')

      if (indexOfColon >= 0) {
        const /** @type {string=} */ ns = NAMESPACE_MAP[attribute.slice(0, indexOfColon)]

        if (ns != null) {
          const name = attribute.slice(indexOfColon + 1)

          if (value === true) {
            n.setAttributeNS(ns, name, '')
          } else if (value === false) {
            n.removeAttributeNS(ns, name)
          } else {
            // @ts-expect-error: TS2345
            n.setAttributeNS(ns, name, value)
          }

          continue
        }
      }

      if (value === true) {
        n.setAttribute(attribute, '')
      } else if (value === false) {
        n.removeAttribute(attribute)
      } else {
        // @ts-expect-error: TS2345
        n.setAttribute(attribute, value)
      }
    }

    for (const property in p) {
      const value = p[property]
      if (property in n && isObject(value)) {
        for (const key2 in value) {
          // @ts-expect-error: TS7053
          n[property][key2] = value[key2]
        }
      } else {
        // @ts-expect-error: TS7053
        n[property] = value
      }
    }

    for (const item of i) {
      const node = isObject(item)
        // @ts-expect-error: TS2769
        ? $(item, namespace)
        : isString(item) || isNumber(item)
          // @ts-expect-error: TS2345
          ? document.createTextNode(item)
          : item

      n.appendChild(node)
    }
  }

  // @ts-expect-error: TS2322
  return n
}

/**
 * A helper for creating `<symbol>` elements.
 */
export const $symbol = (/** @type {$Config | string} */ config) => {
  if (isString(config)) {
    config = JSON.parse(config)
  }
  // @ts-expect-error: TS2698
  return $({ ...config, t: 'symbol' }, NAMESPACE_SVG)
}

/**
 * A helper for creating template `<svg>` elements.
 */
export const $def = (/** @type {($Config | string)[]} */ configs) => $({
  t: 'div',
  p: {
    style: { display: 'none' }
  },
  i: configs.map($symbol)
})

/**
 * A helper for creating `<svg><use>` elements.
 *
 * @returns {$Config & TConfig<'svg'>}
 */
export const $use = (/** @type {string} */ id) => ({
  t: 'svg',
  i: [{
    t: 'use', a: { 'xlink:href': '#' + id }
  }]
})

/**
 * A helper for creating `<h1>` elements.
 */
export const $h1 = (/** @type {NonNullable<$Config['i']>} */ ...i) => $({ t: 'h1', i })

/**
 * A helper for creating `<h2>` elements.
 */
export const $h2 = (/** @type {NonNullable<$Config['i']>} */ ...i) => $({ t: 'h2', i })

/**
 * A helper for creating `<h3>` elements.
 */
export const $h3 = (/** @type {NonNullable<$Config['i']>} */ ...i) => $({ t: 'h3', i })

/**
 * A helper for creating `<h4>` elements.
 */
export const $h4 = (/** @type {NonNullable<$Config['i']>} */ ...i) => $({ t: 'h4', i })

/**
 * A helper for creating `<h5>` elements.
 */
export const $h5 = (/** @type {NonNullable<$Config['i']>} */ ...i) => $({ t: 'h5', i })

/**
 * A helper for creating `<h6>` elements.
 */
export const $h6 = (/** @type {NonNullable<$Config['i']>} */ ...i) => $({ t: 'h6', i })

/**
 * A helper for creating `<img>` elements.
 */
export const $img = (/** @type {string} */ src) => $({ t: 'img', a: { src } })

/**
 * A helper for creating`<p>` elements.
 */
export const $p = (/** @type {NonNullable<$Config['i']>} */ ...i) => $({ t: 'p', i })

/**
 * A helper for creating `<ul>` elements.
 */
export const $ul = (/** @type {NonNullable<$Config['i']>} */ ...items) => $({
  t: 'ul', i: items.map(item => ({ t: 'li', i: [item] }))
})

/**
 * A helper for creating `<a>` elements.
 */
export const $a = (/** @type {string} */ href, /** @type {string?} */ text) => $({
  t: 'a', a: { href }, i: text != null ? [text] : [{ t: 'code', i: [href] }]
})

/**
 * The `$CSS` format provides a hierarchical description of CSS rules.
 *
 * - Keys of sub-objects whose values are NOT objects are treated as CSS attribute, and values are treated as values of those CSS attributes; the concatenation of keys of all parent objects is a CSS rule.
 * - All keys ignore the part starting with a splitter (default: `$$`) sign until the end of the key (e.g. `src$$1` → `src`, `@font-face$$1` → `@font-face`).
 * - In keys specifying CSS attribute, all uppercase letters are replaced by lowercase letters with an additional `-` character preceding them (e.g. `fontFamily` → `font-family`).
 * - Commas in keys that makes a CSS rule cause it to “split” and create separate rules for each part (e.g. `{div:{margin:1,'.a,.b,.c':{margin:2}}}` → `div{margin:1}div.a,div.b,div.c{margin:2}`).
 * - Top-level keys that begin with `@` are not concatenated with sub-object keys.
 *
 * @typedef {{
 *   [ruleOrAttribute: string]: string | number | $CSS;
 * }} $CSS
 */

const _css = (
  /** @type {string | number | $CSS} */ style,
  /** @type {string} */ prefix,
  /** @type {string[]} */ result,
  /** @type {(text: string) => string} */ split
) => {
  // @ts-expect-error: TS2322
  const /** @type {[$CSS | string[], string][]} */ queue = [[style, prefix]]

  while (queue.length > 0) {
    // @ts-expect-error: TS2322
    const /** @type {[$CSS | string[], string]} */ [style2, prefix2] = queue.shift()

    if (Array.isArray(style2)) {
      result.push(
        prefix2,
        prefix2 !== '' ? '{' : '',
        style2.join(';'),
        prefix2 !== '' ? '}' : ''
      )
    } else {
      const /** @type {[$CSS | string[], string][]} */ todo = []
      let arrPushed = false
      let /** @type {string[]} */ arr = []

      for (const key in style2) {
        const value = style2[key]

        if (isString(value) || isNumber(value)) {
          if (!arrPushed) {
            arrPushed = true
            arr = []
            todo.push([arr, prefix2])
          }

          arr.push(`${split(key).replace(/([A-Z])/g, (_, letter) => '-' + letter.toLowerCase())}:${value}`)
        } else {
          arrPushed = false

          const /** @type {string[]} */ newPrefix = []
          const keySplitted = key.split(',')

          for (const prefixItem of prefix2.split(',')) {
            for (const keyItem of keySplitted) {
              newPrefix.push(prefixItem + keyItem)
            }
          }

          todo.push([value, newPrefix.join(',')])
        }
      }

      queue.unshift(...todo)
    }
  }
}

/**
 * CSS-in-JS helper based on the `$CSS` format.
 * An iterative implementation that does not cause a stack overflow exception.
 */
export const $css = (/** @type {$CSS} */ style, /** @type {string} */ splitter = '$$') => {
  const split = (/** @type {string} */ text) => text.split(splitter)[0]
  const /** @type {string[]} */ result = []

  for (const key in style) {
    if (key[0] === '@') {
      result.push(split(key) + '{')
      _css(style[key], '', result, split)
      result.push('}')
    } else {
      _css(style[key], split(key), result, split)
    }
  }

  return result.join('')
}

/**
 * A helper for creating `<style>` elements.
 */
export const $style = (/** @type {$CSS} */ style, /** @type {string=} */ splitter) => $({
  t: 'style',
  i: [$css(style, splitter)]
})

/**
 * A generic helper for syntax highlighting.
 */
export const $highlight = (
  /** @type {RegExp} */ pattern,
  /** @type {any[]}  */ contexts,
  /** @type {(chunk: string, context: any) => string} */ highlight,
  /** @type {string} */ text
) => text
  .replace(/&/g, '&amp;')
  .replace(/</g, '&lt;')
  .replace(/>/g, '&gt;')
  .replace(pattern, (_0, ...matches) => matches.map((chunk, index) => isString(chunk) ? highlight(chunk, contexts[index]) : chunk).find(isString) ?? '')

/**
 * A helper for highlighting JavaScript syntax.
 */
export const $highlightJs = $highlight.bind(0,
  /('.*?'|".*?"|`[\s\S]*?`)|(\/\/.*?\n|\/\*[\s\S]*?\*\/)|((?:break|case|catch|continue|debugger|default|delete|do|else|export|extends|finally|goto|if|import|in|instanceof|is|keyof|new|package|return|switch|throw|try|typeof|void|while|with|yield)(?!\w)|&lt;|&gt;|&amp;|[=.:?!~|*/%+-])|((?:const|eval|function|let|type |var))|(super|this)(?!\w)|((?:false|NaN|null|true|undefined)(?!\w)|\d+(?:\.\d+)?(?:e[+-]?\d+)?)|([$\w]+)(?=\()|([$\w]+)/g,
  ['#ed7', '#776', '#f27', '#6de', '#f92', '#a8f', '#ad3', '#ffe'],
  (chunk, context) => `<span style="color:${context}">${chunk}</span>`
)

/** @returns {string} */
const _locale = (
  /** @type {Record<string, Record<string, string | Record<string, string>>>} */ locales,
  /** @type {string} */ language,
  /** @type {string} */ text,
  /** @type {string} */ version
) => get(locales, language, version, text) ?? get(locales, language, text) ?? text

/**
 * Language translations helper.
 */
export const $locale = (
  /** @type {Record<string, Record<string, string | Record<string, string>>>} */ locales,
  /** @type {string} */ defaultLanguage,
  /** @type {readonly string[]} */ languages = navigator.languages
) => {
  for (const language of languages) {
    if (hasOwnProperty(locales, language)) {
      return _locale.bind(0, locales, language)
    }
  }

  return _locale.bind(0, locales, defaultLanguage)
}
